use log::{debug, error, info, trace, warn};
use std::{
    io::Read,
    process::{Child, Command, Stdio},
    time::Instant, path::{PathBuf, Path},
};
use xcb::{randr, x, Xid, XidNew};
use xcb_xembed::embedder::Embedder;

use bscreensaver_command::{BCommand, create_command_window};
use bscreensaver_util::{*, dialog::DialogExitStatus, settings::Configuration};

use crate::{
    monitor::*,
    pidfd::{CreatePidFd, PidFd},
};

#[derive(Debug, Clone, Copy, PartialEq, PartialOrd)]
pub enum BlankerState {
    Idle = 0,
    Blanked = 1,
    Locked = 2,
}

struct UnlockDialog<'a> {
    monitor_geom: x::Rectangle,
    blanker_window: x::Window,
    embedder: Embedder<'a>,
    event_to_forward: Option<xcb::Event>,
    child: Child,
    child_pidfd: PidFd,
}

impl<'a> Drop for UnlockDialog<'a> {
    fn drop(&mut self) {
        let _ = self.child.kill();
        let _ = self.child.try_wait();
    }
}

type CommandHandler = dyn Fn(&mut Screensaver, Option<&x::ClientMessageEvent>) -> anyhow::Result<()>;

pub struct CommandHandlers<'a> {
    pub restart_handler: &'a CommandHandler,
    pub exit_handler: &'a CommandHandler,
}

struct CommandAtoms {
    blank: x::Atom,
    lock: x::Atom,
    deactivate: x::Atom,
    restart: x::Atom,
    exit: x::Atom,
}

pub struct Screensaver<'a> {
    config: &'a Configuration,
    helper_dir: PathBuf,
    conn: &'a xcb::Connection,
    command_atoms: CommandAtoms,
    command_handlers: &'a CommandHandlers<'a>,
    blanker_state: BlankerState,
    monitors: Vec<Monitor<'a>>,
    last_user_activity: Instant,
    unlock_dialog: Option<UnlockDialog<'a>>,
}

impl<'a> Screensaver<'a> {
    pub fn new<P: AsRef<Path>>(config: &'a Configuration, helper_dir: P, command_handlers: &'a CommandHandlers<'a>, conn: &'a xcb::Connection, screen: &x::Screen) -> anyhow::Result<Self> {
        let _ = create_command_window(conn, screen)?;
        let command_atoms = CommandAtoms {
            blank: create_atom(&conn, BCommand::Blank.atom_name())?,
            lock: create_atom(&conn, BCommand::Lock.atom_name())?,
            deactivate: create_atom(&conn, BCommand::Deactivate.atom_name())?,
            restart: create_atom(&conn, BCommand::Restart.atom_name())?,
            exit: create_atom(&conn, BCommand::Exit.atom_name())?,
        };

        Ok(Self {
            config,
            helper_dir: helper_dir.as_ref().to_path_buf(),
            conn,
            command_atoms,
            command_handlers,
            blanker_state: BlankerState::Idle,
            monitors: Monitor::set_up_all(conn)?,
            last_user_activity: Instant::now(),
            unlock_dialog: None,
        })
    }

    pub fn blanker_state(&self) -> BlankerState {
        self.blanker_state
    }

    pub fn last_user_activity(&self) -> Instant {
        self.last_user_activity
    }

    pub fn xcb_connection(&self) -> &xcb::Connection {
        self.conn
    }

    pub fn unlock_dialog_pidfd(&self) -> Option<&PidFd> {
        self.unlock_dialog.as_ref().map(|ud| &ud.child_pidfd)
    }

    pub fn handle_xcb_events(&mut self) -> anyhow::Result<()> {
        loop {
            if let Some(event) = self.conn.poll_for_event()? {
                let embedder_handled = if let Some(mut unlock_dialog) = self.unlock_dialog.take() {
                    match unlock_dialog.embedder.event(&event) {
                        Err(err) => {
                            // XXX: should we assume unlock dialog is dead here?
                            warn!("Error sending event to unlock dialog: {}", err);
                            false
                        },
                        Ok(handled) => {
                            self.unlock_dialog = Some(unlock_dialog);
                            handled
                        },
                    }
                } else {
                    false
                };

                // Allow the code below to handle key presses regardless of whether or not
                // the embedder handled it.
                let embedder_handled = match event {
                    xcb::Event::X(x::Event::KeyPress(_)) => false,
                    _ => embedder_handled,
                };

                if !embedder_handled {
                    match event {
                        xcb::Event::RandR(randr::Event::Notify(ev)) => {
                            debug!("Got xrandr notify event: {:#?}", ev);
                            self.monitors = Monitor::set_up_all(self.conn)?;
                            match self.blanker_state {
                                BlankerState::Idle => (),
                                BlankerState::Blanked => {
                                    self.blanker_state = BlankerState::Idle;
                                    self.blank_screen()?;
                                },
                                BlankerState::Locked => {
                                    self.blanker_state = BlankerState::Idle;
                                    self.lock_screen()?;
                                },
                            }
                        },
                        xcb::Event::Input(_) => {
                            // TODO: implement some sort of hysteresis/debouncing for mouse motion
                            self.last_user_activity = Instant::now();
                        },
                        xcb::Event::X(x::Event::ClientMessage(ev)) => {
                            let res = match ev.r#type() {
                                b if b == self.command_atoms.blank => Some(self.blank_screen()),
                                l if l == self.command_atoms.lock => Some(self.lock_screen()),
                                d if d == self.command_atoms.deactivate => {
                                    self.last_user_activity = Instant::now();
                                    match self.blanker_state {
                                        BlankerState::Idle => Some(Ok(())),
                                        BlankerState::Blanked => Some(self.unblank_screen()),
                                        BlankerState::Locked => {
                                            if self.unlock_dialog.is_none() {
                                                match self.start_unlock_dialog(None) {
                                                    Ok(unlock_dialog) => {
                                                        self.unlock_dialog = Some(unlock_dialog);
                                                        Some(Ok(()))
                                                    },
                                                    Err(err) => Some(Err(err)),
                                                }
                                            } else {
                                                Some(Ok(()))
                                            }
                                        },
                                    }
                                },
                                r if r == self.command_atoms.restart => Some((self.command_handlers.restart_handler)(self, Some(&ev))),
                                e if e == self.command_atoms.exit => Some((self.command_handlers.exit_handler)(self, Some(&ev))),
                                _ => None,
                            };

                            if let Some(res) = res {
                                let is_success = if let Err(err) = res {
                                    warn!("Failed to handle remote command {}: {}", ev.r#type().resource_id(), err);
                                    false
                                } else {
                                    true
                                };
                                crate::send_command_reply(self.conn, &ev, is_success);
                            }
                        },
                        xcb::Event::X(x::Event::MapNotify(ev)) if ev.window() == self.unlock_dialog_window() => {
                            debug!("Unlock dialog mapped, requesting focus");
                            if let Some(ref mut unlock_dialog) = &mut self.unlock_dialog {
                                let _ = unlock_dialog.embedder.activate_client();
                                if let Err(err) = unlock_dialog.embedder.focus_client(xcb_xembed::XEmbedFocus::Current) {
                                    warn!("Failed to focus unlock dialog: {}", err);
                                }
                                if let Some(event_to_forward) = unlock_dialog.event_to_forward.take() {
                                    let _ = unlock_dialog.embedder.event(&event_to_forward);
                                }
                            }
                        },
                        xcb::Event::X(x::Event::ConfigureNotify(ev)) if ev.window() == self.embedder_window() => {
                            if let Some(unlock_dialog) = &self.unlock_dialog {
                                let monitor_geom = &unlock_dialog.monitor_geom;
                                let x = std::cmp::max(0, monitor_geom.x as i32 + monitor_geom.width as i32 / 2 - ev.width() as i32 / 2);
                                let y = std::cmp::max(0, monitor_geom.y as i32 + monitor_geom.height as i32 / 2 - ev.height() as i32 / 2);
                                if x != ev.x() as i32 || y != ev.y() as i32 {
                                    self.conn.send_and_check_request(&x::ConfigureWindow {
                                        window: unlock_dialog.embedder.embedder_window(),
                                        value_list: &[
                                            x::ConfigWindow::X(x),
                                            x::ConfigWindow::Y(y),
                                        ],
                                    })?;
                                }
                            }
                        },
                        xcb::Event::X(x::Event::Expose(ev)) => {
                            if let Some(monitor) = self.monitors.iter().find(|m| ev.window() == m.blanker_window || ev.window() == m.unlock_window) {
                                debug!("got expose for {} at {}x{}+{}+{}", ev.window().resource_id(), ev.width(), ev.height(), ev.x(), ev.y());
                                self.conn.send_and_check_request(&x::PolyFillRectangle {
                                    drawable: x::Drawable::Window(ev.window()),
                                    gc: monitor.black_gc,
                                    rectangles: &[
                                        x::Rectangle {
                                            x: ev.x() as i16,
                                            y: ev.y() as i16,
                                            width: ev.width(),
                                            height: ev.height(),
                                        },
                                    ],
                                })?;
                            }
                        },
                        ev @ xcb::Event::X(x::Event::MotionNotify(_)) => self.handle_user_activity(ev)?,
                        xcb::Event::X(x::Event::KeyPress(ev)) => {
                            if let Err(err) = keysym_for_keypress(self.conn, &ev).and_then(|keysym| match keysym {
                                Some(keysym) if keysym == xkb::key::XF86MonBrightnessUp && self.config.handle_brightness_keys => self.monitors.iter().map(|monitor| monitor.brightness_up()).collect(),
                                Some(keysym) if keysym == xkb::key::XF86MonBrightnessDown && self.config.handle_brightness_keys => self.monitors.iter().map(|monitor| monitor.brightness_down()).collect(),
                                _ => Ok(()),
                            }) {
                                warn!("Failed to handle key press {}: {}", ev.detail(), err);
                            }
                            self.handle_user_activity(xcb::Event::X(x::Event::KeyPress(ev)))?;
                        },
                        ev => trace!("Got other event: {:#?}", ev),
                    }
                }
            } else {
                break;
            }
        }

        Ok(())
    }

    fn handle_user_activity(&mut self, ev: xcb::Event) -> anyhow::Result<()> {
        match self.blanker_state {
            BlankerState::Idle => Ok(()),
            BlankerState::Blanked => self.unblank_screen(),
            BlankerState::Locked => match &self.unlock_dialog {
                None => {
                    self.unlock_dialog = match self.start_unlock_dialog(Some(ev)) {
                        Err(err) => {
                            error!("Unable to start unlock dialog: {}", err);
                            None
                        },
                        Ok(unlock_dialog) => Some(unlock_dialog),
                    };
                    Ok(())
                },
                Some(unlock_dialog) => {
                    let mut cookies = Vec::new();
                    for win in [unlock_dialog.blanker_window, unlock_dialog.embedder.embedder_window(), unlock_dialog.embedder.client_window()] {
                        cookies.push(self.conn.send_request_checked(&x::ConfigureWindow {
                            window: win,
                            value_list: &[
                                x::ConfigWindow::StackMode(x::StackMode::Above),
                            ],
                        }));
                    }
                    for cookie in cookies {
                        self.conn.check_request(cookie)?;
                    }
                    Ok(())
                },
            }
        }
    }

    fn embedder_window(&self) -> x::Window {
        self.unlock_dialog.as_ref().map(|ud| ud.embedder.embedder_window()).unwrap_or(x::WINDOW_NONE)
    }

    fn unlock_dialog_window(&self) -> x::Window {
        self.unlock_dialog.as_ref().map(|ud| ud.embedder.client_window()).unwrap_or(x::WINDOW_NONE)
    }

    pub fn handle_unlock_dialog_quit(&mut self) -> anyhow::Result<()> {
        if let Some(mut unlock_dialog) = self.unlock_dialog.take() {
            match unlock_dialog.child.try_wait() {
                Err(err) => {
                    warn!("Failed to check unlock dialog's status: {}", err);
                    self.hide_cursor();
                },
                Ok(Some(status)) => {
                    let ok = match DialogExitStatus::try_from(status) {
                        Ok(result) => {
                            info!("{}", result);
                            result == DialogExitStatus::AuthSucceeded
                        },
                        Err(error) => {
                            warn!("{}", error);
                            false
                        },
                    };

                    if ok {
                        self.unlock_screen()?;
                    } else {
                        self.hide_cursor();
                    }
                },
                Ok(None) => self.unlock_dialog = Some(unlock_dialog),  // dialog still running
            }
        }
        Ok(())
    }

    fn start_unlock_dialog(&self, trigger_event: Option<xcb::Event>) -> anyhow::Result<UnlockDialog<'a>> {
        let mut monitor_data = None;
        for monitor in &self.monitors {
            let cookie = self.conn.send_request(&x::QueryPointer {
                window: monitor.root,
            });
            let reply = self.conn.wait_for_reply(cookie)?;
            let px = reply.root_x() as i32;
            let py = reply.root_y() as i32;
            if reply.same_screen()
                && px >= monitor.x as i32 && px < monitor.x as i32 + monitor.width as i32
                    && py >= monitor.y as i32 && py < monitor.y as i32 + monitor.height as i32
            {
                monitor_data = Some((monitor.geometry(), monitor.blanker_window, monitor.unlock_window));
                break;
            }
        }
        let (monitor_geom, blanker_window, unlock_window) = monitor_data.unwrap_or_else(|| {
            warn!("Unable to determine which monitor pointer is on; using first one");
            let monitor = self.monitors.iter().nth(0).unwrap();
            (monitor.geometry(), monitor.blanker_window, monitor.unlock_window)
        });

        for monitor in &self.monitors {
            monitor.show_cursor();
        }

        let trigger_event = match trigger_event {
            Some(xcb::Event::X(x::Event::KeyPress(ev))) => match keysym_for_keypress(self.conn, &ev) {
                Err(err) => {
                    warn!("Failed to get keysym for key press event: {}", err);
                    Some(xcb::Event::X(x::Event::KeyPress(ev)))
                },
                Ok(Some(keysym)) if keysym == xkb::key::KP_Enter || keysym == xkb::key::ISO_Enter || keysym == xkb::key::Return || keysym == xkb::key::Escape =>
                    // don't forward an <enter> or <esc> to the dialog, as that will make it activate/close immediately
                    None,
                _ => Some(xcb::Event::X(x::Event::KeyPress(ev))),
            },
            te => te,
        };

        let mut child = Command::new(self.helper_dir.join(self.config.dialog_backend.binary_name()))
            .stdout(Stdio::piped())
            .spawn()?;

        let child_pidfd = child.create_pidfd()?;
        let mut child_out = child.stdout.take().unwrap();

        let mut xid_buf: [u8; 4] = [0; 4];
        child_out.read_exact(&mut xid_buf)?;
        let client_window = {
            let wid: u32 = ((xid_buf[0] as u32) << 24) | ((xid_buf[1] as u32) << 16) | ((xid_buf[2] as u32) << 8) | (xid_buf[3] as u32);
            unsafe { x::Window::new(wid) }
        };
        debug!("Dialog process created plug window 0x{:x}", client_window.resource_id());

        let cookie = self.conn.send_request(&x::GetWindowAttributes {
            window: client_window,
        });
        let reply = self.conn.wait_for_reply(cookie)?;
        if !reply.your_event_mask().contains(x::EventMask::STRUCTURE_NOTIFY) {
            self.conn.send_and_check_request(&x::ChangeWindowAttributes {
                window: client_window,
                value_list: &[
                    x::Cw::EventMask(reply.your_event_mask() | x::EventMask::STRUCTURE_NOTIFY),
                ],
            })?;
        }

        let embedder = Embedder::start(self.conn, unlock_window, client_window)?;

        Ok(UnlockDialog {
            monitor_geom,
            blanker_window,
            embedder,
            event_to_forward: trigger_event,
            child,
            child_pidfd,
        })
    }

    pub fn stop_unlock_dialog(&mut self) {
        let _ = self.unlock_dialog.take();
    }

    pub fn blank_screen(&mut self) -> anyhow::Result<()> {
        if self.blanker_state >= BlankerState::Blanked {
            return Ok(())
        }

        info!("Blanking");

        for monitor in &self.monitors {
            monitor.blank()?;
        }

        self.blanker_state = BlankerState::Blanked;

        Ok(())
    }

    fn unblank_screen(&mut self) -> anyhow::Result<()> {
        if self.blanker_state != BlankerState::Blanked {
            return Ok(());
        }

        info!("Unblanking");

        for monitor in &self.monitors {
            monitor.unblank()?;
        }

        self.blanker_state = BlankerState::Idle;

        Ok(())
    }

    pub fn lock_screen(&mut self) -> anyhow::Result<()> {
        self.blank_screen()?;
        if self.blanker_state >= BlankerState::Locked {
            return Ok(());
        }

        info!("Locking");

        for monitor in &self.monitors {
            monitor.lock()?;
        }

        self.blanker_state = BlankerState::Locked;

        Ok(())
    }

    fn unlock_screen(&mut self) -> anyhow::Result<()> {
        if self.blanker_state != BlankerState::Locked {
            return Ok(());
        }

        info!("Unlocking");

        for monitor in &self.monitors {
            monitor.unlock()?;
        }

        self.blanker_state = BlankerState::Blanked;
        self.unblank_screen()?;

        Ok(())
    }

    fn hide_cursor(&self) {
        for monitor in &self.monitors {
            monitor.hide_cursor();
        }
    }
}

fn keysym_for_keypress(conn: &xcb::Connection, ev: &x::KeyPressEvent) -> anyhow::Result<Option<xkb::Keysym>> {
    let ctx = xkb::Context::new(xkb::context::Flags::NO_FLAGS);
    let device = xkb::x11::device(conn)
        .map_err(|_| anyhow::anyhow!("Failed to get xkb device"))?;
    let keymap = xkb::x11::keymap(conn, device, &ctx, xkb::keymap::compile::Flags::NO_FLAGS)
        .map_err(|_| anyhow::anyhow!("Failed to get xkb keymap"))?;
    let state = xkb::x11::state(conn, device, &keymap)
        .map_err(|_| anyhow::anyhow!("Failed to get xkb state"))?;
    Ok(state.key(ev.detail()).sym())
}
